from io import BufferedReader
from struct import unpack
from copy import deepcopy
from contextlib import contextmanager

from .structures import Octree, Node, Leaf


class Command:
    def __init__(self, name, count=1, value=None):
        self.command = name
        self.count = count
        self.value = value

    def __str__(self):
        out = self.command
        if self.value is not None:
            out += " " + str(self.value)
        return out


class Builder:
    def __init__(self, level):
        self.level = level
        self.creating_index = 0
        self.end_of_line = True
        self.artefact:"list[Builder]" = []

    def crnd(self, *args):
        if self.end_of_line == False:
            self.artefact[self.creating_index].crnd()
        else:
            self.artefact.append(Builder(self.level-1))
            self.end_of_line = False

    def flnd(self, data):
        if self.end_of_line == False:
            if self.artefact[self.creating_index].flnd(data):
                self.creating_index += 1
                self.end_of_line = True
        else:
            self.artefact.append(Leaf(self.level -1, data))
            self.creating_index += 1
        
        if self.creating_index == 8:
            self.contents = self.artefact
            self.__class__ = Node
            del self.artefact
            del self.end_of_line
            del self.creating_index
            return True


class BuilderHelper(Builder):
    class VersionError(Exception):
        def __init__(self, given, required):
            self.message = f"Version {required} was required, but {given} was given"
            super().__init__(self.message)
    
    
    def __init__(self):
        pass


    def header(self, value):
        if value != "0.0.2":
            raise self.VersionError(value, "0.0.2")


    def seed(self, value):
        self.octree = Octree(value)
        super().__init__(value + 1)


    def route(self, command:Command):
        for _ in range(command.count):
            getattr(self, command.command)(deepcopy(command.value))


    @contextmanager
    def build(self):
        try:
            yield self
        finally:
            self.octree.octree = self.artefact[0]


class LoadingStream:
    commands = {
        b"\x00":"header", b"\x01":"seed",
        
        b"\x04":"crnd", b"\x07":"flnd", 
        b"\x08":"nlnd", b"\x09":"vdnd", b"\x0a":"fsnd", b"\x0b":"trnd",
        
        b"\x20":  "i8", b"\x21": "i16", b"\x22": "i32", b"\x23": "i64",
        b"\x24":  "u8", b"\x25": "u16", b"\x26": "u32", b"\x27": "u64",
        b"\x28": "f32", b"\x29": "f64", b"\x2a": "c64", b"\x2b":"c128",
        b"\x40": "Str", b"\x41":"List", b"\x42":"Dict", b"\x43":"Set"
    }


    def __init__(self, io:BufferedReader) -> None:
        self.io = io


    def convert(self):
        return getattr(self, self.commands[self.read()])()
        

    def read(self) -> bytes:
        if (next_byte := self.io.read(1)):
            return next_byte
        else:
            raise EOFError


    def i8(self):
        return int.from_bytes(self.io.read(1), byteorder="big", signed=True)

    def i16(self):
        return int.from_bytes(self.io.read(2), byteorder="big", signed=True)

    def i32(self):
        return int.from_bytes(self.io.read(4), byteorder="big", signed=True)

    def i64(self):
        return int.from_bytes(self.io.read(8), byteorder="big", signed=True)

    def u8(self):
        return int.from_bytes(self.io.read(1), byteorder="big", signed=False)

    def u16(self):
        return int.from_bytes(self.io.read(2), byteorder="big", signed=False)

    def u32(self):
        return int.from_bytes(self.io.read(4), byteorder="big", signed=False)

    def u64(self):
        return int.from_bytes(self.io.read(8), byteorder="big", signed=False)

    def f32(self):
        return unpack("f", self.io.read(4))

    def f64(self):
        return unpack("d", self.io.read(8))

    def c64(self):
        return complex(*unpack("ff", self.io.read(8)))

    def c128(self):
        return complex(*unpack("dd", self.io.read(16)))

    def Str(self):
        return self.io.read(self.u16()).decode()

    def List(self):
        return list(self.convert() for _ in range(self.u16()))

    def Dict(self):
        return dict((self.convert(), self.convert()) for _ in range(self.u16()))

    def Set(self):
        return set(self.convert() for _ in range(self.u16()))


class Loader:
    commands = {
        b"\x00":"header", b"\x01":"seed",
        
        b"\x04":"crnd", b"\x07":"flnd", 
        b"\x08":"nlnd", b"\x09":"vdnd", b"\x0a":"fsnd", b"\x0b":"trnd",
    }


    def __init__(self, io:BufferedReader, factory) -> None:
        self.factory = factory
        self.converter = LoadingStream(io)


    def __iter__(self):
        return self


    def __next__(self):
        try:
            return getattr(self, self.commands[self.converter.read()])()
        except EOFError:
            raise StopIteration


    def header(self):
        return Command("header", value=self.converter.convert())
    
    def seed(self):
        return Command("seed", value=self.converter.convert())
    
    def crnd(self):
        return Command("crnd", self.converter.u8())

    def flnd(self):
        return Command("flnd", self.converter.u8(), self.factory.from_file(self.factory, self.converter)) # here custom from_file method should be inserted

        
    def nlnd(self):
        return Command("flnd", self.converter.u8())

    def vdnd(self):
        return Command("flnd", self.converter.u8(), value=Ellipsis)

    def fsnd(self):
        return Command("flnd", self.converter.u8(), value=False)

    def trnd(self):
        return Command("flnd", self.converter.u8(), value=True)



import glm


class Geometry():
    @staticmethod
    def index_from_coords(coord_list, level=0):
        return sum(glm.i16vec3(coord_list) * glm.i16vec3(1,2,4) / 2**level)

    @staticmethod   
    def coord_gen(num):
        for i in range(8):
            yield glm.vec3(num) % glm.vec3(2, 4, 8) // glm.vec3(1, 2, 4)

    @staticmethod
    def coords_from_index(num, level=0):
        return (glm.vec3(num) % glm.vec3(2, 4, 8) // glm.vec3(1, 2, 4) * 2**level).to_list()
        
    @staticmethod
    def coord_addition(coord1, coord2):
        return (glm.vec3(coord1) + glm.vec3(coord2)).to_list()

    @staticmethod
    def coord_div(coord, denom):
        return (glm.vec3(coord) // denom).to_list()

    @staticmethod
    def coord_mod(coord, denom):
        return (glm.vec3(coord) % denom).to_list()

    @staticmethod
    def coord_compare_less(coord, value):
        if 1 in (glm.i16vec3(coord) < glm.i16vec3(value)):
            return False
        else:
            return True


from contextlib import contextmanager
from copy import deepcopy
import math

from .util import Geometry


class Leaf:
    def __init__(self, level, data):
        self.value = data
        self.level = level
        if self.level < 0:
            raise ValueError(self.level)


    def clone(self, level):
        return Leaf(self.level, deepcopy(self.value))


    def get(self, coords):
        return self.value


    def set(self, coords, level) -> "Leaf | Node":
        if self.level == level:
            return self
        else:
            self.subdivide()
            return self.set(coords, level)


    def subdivide(self):
        self.__class__ = Node
        self.contents = list(Leaf(self.level-1, deepcopy(self.value)) for i in range(8))
        del self.value


    def make_leaf(self, data):
        """sets value for leaf"""
        self.value = data


    def make_node(self, node):
        if node.__class__ == Leaf:
            self.value = deepcopy(node.value)
        elif node.__class__ == Node:
            self.__class__ = Node
            self.contents = list(member.clone(self.level-1) for member in node.contents)
            del self.value


    def defragment(self):...


    def __str__(self):
        return str({"level":self.level, "data":str(self.value)})


    def __bool__(self):
        return bool(self.value) 


    def __iter__(self):
        self.has_returned = False
        return self


    def __next__(self):
        if self.has_returned:
            raise StopIteration
        else:
            self.has_returned = True
            return {"coords": (0, 0, 0), "level":self.level, "void":bool(self), "data":self.value}


    def __len__(self):
        return 1


    def __eq__(self, o: "Leaf") -> bool:
        return self.__class__ == o.__class__ and self.value == o.value


class Node:
    def __init__(self, level):
        self.contents = list(Leaf(level-1, None) for i in range(8))
        self.level = level
        if self.level < 0:
            raise ValueError(self.level)


    def clone(self, level):
        obj = Node(level)
        obj.contents = list(member.clone(obj.level-1) for member in self.contents)
        return obj


    # Navigation Methods
    def get(self, coords):
        next_coords = Geometry.coord_mod(coords, 2**self.level)
        next_index = Geometry.index_from_coords(Geometry.coord_div(next_coords, 2**(self.level-1)))
        return self.contents[next_index].get(next_coords)
        

    def set(self, coords, level) -> "Leaf | Node":
        if self.level == level:
            return self
        else:
            next_coords = Geometry.coord_mod(coords, 2**self.level)
            next_index = Geometry.index_from_coords(
                Geometry.coord_div(
                    next_coords,
                    2**(self.level-1)
                )
            )
            return self.contents[next_index].set(next_coords, level)


    def subdivide(self):
        self.contents = list(self.clone(self.level-1) for _ in range(8))


    def make_leaf(self, data=None):
        self.__class__ = Leaf
        self.value = data
        del self.contents


    def make_node(self, node):
        if node.__class__ == Leaf:
            self.__class__ = Leaf
            self.value = deepcopy(node.value)
        elif node.__class__ == Node:
            self.__class__ = Node
            self.contents = list(member.clone(self.level-1) for member in node.contents)


    def defragment(self):
        for node in self.contents:
            node.defragment()
        if all(self.contents[i] == self.contents[(i+1)%8] for i in range(8)):
            self.make_leaf(self.contents[0].value)


    def __str__(self):
        return "\n".join(str(i) for i in self.contents)


    def __iter__(self):
        self.n = 0
        self.reading = iter(self.contents[0])
        return self


    def __next__(self):
        try:
            out = next(self.reading)
        except StopIteration:
            self.n += 1
            if self.n == 8:
                raise StopIteration
            self.reading = iter(self.contents[self.n])
            out = next(self.reading)
        out["coords"] = tuple(a + b for a, b in zip(Geometry.coords_from_index(self.n, self.level - 1), out["coords"]))
        return out


    def __len__(self):
        return sum(len(child) for child in self)


    def __eq__(self, o: "Leaf | Node"):
        return self.__class__ == o.__class__ and all(a == b for a, b in zip(self.contents, o.contents))


class Octree:
    def __init__(self, level, data=None):
        self.octree:"Leaf | Node" = Leaf(level, data)
        self.level = level


    def _get(self, coords):
        if any(0 > value >= 2**self.level for value in coords):
            raise IndexError()
        else:
            return self.octree.get(coords)


    def _set(self, coords, level) -> "Leaf | Node":
        if any(0 > value >= 2**self.level for value in coords):
            print("out of range")
            raise IndexError()
        return self.octree.set(coords, level)


    def _raise(self):
        print("raising")
        new_octree = Node(self.level + 1, 0, self)
        new_octree.contents = list(Leaf(self.level, None) for _ in range(8))
        new_octree.contents[0] = self.octree
        self.octree = new_octree


    @contextmanager
    def edit(self):
        try:
            yield self
        finally:
            self.octree.defragment()

    # WIP
    def get_sub_tree(self, coords: "tuple", level) -> "Octree":
        """-WIP- Clone a segment of the octree to a new object"""
        node = self._set(coords, level)
        obj = Octree(node.level)
        obj.octree = node.clone(obj.level)
        return obj


    def set_sub_tree(self, coords: "tuple", level, src: "Octree"):
        """-WIP- Insert a seperate Octree to the designated location"""
        self._set(coords, level).make_node(src.octree)


    # spec methods
    def __str__(self):
        return str(self.octree)


    def __iter__(self):
        return iter(self.octree)
    

    def __getitem__(self, index: "tuple"):
        if len(index) != 3:
            raise TypeError("octree key must be a 3-tuple (x, y, z)")
        return self._get(index) 


    def __setitem__(self, index: "tuple", data):
        if len(index) == 3:
            index = (*index, 0)
        if len(index) != 4:
            raise TypeError("octree key must be a 3- or 4-tuple (x, y, z[, level])")
        self._set(index[:3], index[3]).make_leaf(data)


    def __len__(self):
        return len(self.octree)


    def __eq__(self, o: "Octree") -> bool:
        return self.__class__ == o.__class__ and self.octree == o.octree

    
    def __ne__(self, o: "Octree") -> bool:
        return not self == o


from .structures import Octree
from .visualiser import Display
from .reader import LoadingStream
from .saveing import SavingStream
from .interface import Interface


class _Default(Interface):
    def from_file(cls, loader:LoadingStream):
        return loader.convert()

    def to_file(obj, saver:SavingStream):
        saver.convert(obj) 


def load(file_path):
    return _Default.load(file_path)


def save(octree, file_path):
    _Default.save(octree, file_path)


import abc
from typing import final

from .structures import Octree
from .reader import Loader, LoadingStream, BuilderHelper
from .saveing import Saver, SavingStream, scanner


class Interface(abc.ABC):
    @classmethod
    @final
    def load(cls, file_path) -> Octree:
        with open(file_path, "rb") as io:
            build_help = BuilderHelper()
            with build_help.build():
                for command in Loader(io, cls):
                    build_help.route(command)
            return build_help.octree        


    @classmethod
    @final
    def save(cls, octree:Octree, file_path):
        with open(file_path, "wb") as io:
            saver = Saver(io, cls)
            for command in scanner(octree):
                saver.translate(command)


    @classmethod
    @abc.abstractmethod
    def from_file(cls, loader:LoadingStream):...


    @staticmethod
    @abc.abstractmethod
    def to_file(obj, saver:SavingStream):...



import math, sys, struct
import glfw
import OpenGL.GL as gl
import numpy
import glm
import ctypes

cs = """
    #version 430 core

    uniform uint vertexcount;

    struct DrawCommand {
        uint count;
        uint instanceCount;
        uint firstIndex;
        uint baseVertex;
        uint baseInstance;
    };

    layout(local_size_x = 1) in;

    layout(std140, binding = 0) uniform barrierData {
        vec3 direction;
        vec3 barrier;
    };

    layout(std140, binding = 1) buffer modelMatrix {
        mat4 model[];
    };

    layout(std430, binding = 2) buffer drawIndirectBuffer {
        DrawCommand commands[];
    };


    void main(){
        const uint index = gl_WorkGroupID.x; // this shader may only be dispatched in the x dimension

        bool draw = dot((model[index] * vec4(0.5, 0.5, 0.5, 1.0)).xyz - barrier, direction) > 0.0;

        commands[index] = DrawCommand(
            vertexcount,
            draw ? 1 : 0,
            0,
            0,
            index
        );        
    }
    """

vs = """
    #version 430 core
    
    layout(location = 0) in vec3 vertexPosition_modelspace;
    layout(location = 2) in mat4 model;

    layout(std140, binding = 3) uniform matricies {
        mat4 projection;
        mat4 view;
    };

    uniform vec3 staticcolor;

    out vec3 fragmentColor;
    
    void main(){
        mat4 fullMatrix = projection * view * model;
        gl_Position = fullMatrix * vec4(vertexPosition_modelspace, 1.0);
        fragmentColor = staticcolor;
    }
    """

fs = """
    #version 430 core
    in vec3 fragmentColor;

    out vec3 color;

    void main(){
        color = fragmentColor;
    }
    """

indexed_vertecies = [
        0.0, 0.0, 0.0,
        0.0, 0.0, 1.0,
        0.0, 1.0, 0.0,
        0.0, 1.0, 1.0,
        1.0, 0.0, 0.0,
        1.0, 0.0, 1.0,
        1.0, 1.0, 0.0,
        1.0, 1.0, 1.0
        ]

cube_indicies = [
        7, 4, 6,
        1, 5, 7, 
        2, 3, 7, 
        1, 7, 3, 
        7, 5, 4,
        2, 7, 6,
        4, 2, 6,
        0, 1, 3,
        3, 2, 0,
        0, 5, 1,
        0, 4, 5,
        0, 2, 4
        ]

frame_indicies = [
        0, 1,
        1, 3,
        1, 5,
        6, 7,
        0, 2,
        2, 3, 
        2, 6, 
        5, 7, 
        0, 4,
        4, 5, 
        4, 6,
        3, 7
        ]


def compile_compute_shader(cs):
    compute_shader_id = gl.glCreateShader(gl.GL_COMPUTE_SHADER)

    gl.glShaderSource(compute_shader_id, cs)
    gl.glCompileShader(compute_shader_id)
    if not gl.glGetShaderiv(compute_shader_id, gl.GL_COMPILE_STATUS):
        print("compute shader compile error:\n\t", "\n\t".join(gl.glGetShaderInfoLog(compute_shader_id).decode().split("\n")), sep="")

    program_id = gl.glCreateProgram()
    gl.glAttachShader(program_id, compute_shader_id)

    gl.glLinkProgram(program_id)
    if not gl.glGetProgramiv(program_id, gl.GL_LINK_STATUS):
        print("shader link error:\n\t", "\n\t".join(gl.glGetProgramInfoLog(program_id).decode().split("\n")), sep="")
        sys.exit()

    gl.glDetachShader(program_id, compute_shader_id)
    gl.glDeleteShader(compute_shader_id)

    return program_id


def load_shaders(vs, fs):
    vertex_shader_id = gl.glCreateShader(gl.GL_VERTEX_SHADER)
    fragment_shader_id = gl.glCreateShader(gl.GL_FRAGMENT_SHADER)

    gl.glShaderSource(vertex_shader_id, vs)
    gl.glCompileShader(vertex_shader_id)
    if not gl.glGetShaderiv(vertex_shader_id, gl.GL_COMPILE_STATUS):
        print("vertex shader compile error:\n\t", "\n\t".join(gl.glGetShaderInfoLog(vertex_shader_id).decode().split("\n")), sep="")

    gl.glShaderSource(fragment_shader_id, fs)
    gl.glCompileShader(fragment_shader_id)
    if not gl.glGetShaderiv(fragment_shader_id, gl.GL_COMPILE_STATUS):
        print("fragment shader compile error:\n\t", "\n\t".join(gl.glGetShaderInfoLog(fragment_shader_id).decode().split("\n")), sep="")

    program_id = gl.glCreateProgram()
    gl.glAttachShader(program_id, vertex_shader_id)
    gl.glAttachShader(program_id, fragment_shader_id)

    gl.glLinkProgram(program_id)
    if not gl.glGetProgramiv(program_id, gl.GL_LINK_STATUS):
        print("shader link error:\n\t", "\n\t".join(gl.glGetProgramInfoLog(program_id).decode().split("\n")), sep="")
        sys.exit()

    gl.glDetachShader(program_id, vertex_shader_id)
    gl.glDetachShader(program_id, fragment_shader_id)

    gl.glDeleteShader(vertex_shader_id)
    gl.glDeleteShader(fragment_shader_id)

    return program_id


class Display:
    def __init__(self, octree):
        self.octree = octree
        self.window_dims = [1280, 960]
        
        self.fulcrum = glm.vec3(2**(octree.octree.level-1))  # orbit center of camera
        self.direction = glm.vec3(0)
        self.position = glm.vec3(0)
        self.dist = 2**(octree.octree.level+1)
        self.barrier = 0
        
        self.h_angle = math.pi/4
        self.v_angle = 0 # -1*math.pi/8
        self.initial_fov = 60.0

        self.fill_color = glm.vec3(0.9)

        self.speed = 1.0
        self.mouse_speed = 0.5

        self.projectionmatrix = glm.perspective(glm.radians(60.0), 3/2, 0.1, 10000.0)
    
        if not glfw.init():
            print(glfw.get_error())
            return

        self.last_time = glfw.get_time()
        self.last_mouse_x, self.last_mouse_y = None, None

        glfw.window_hint(glfw.SAMPLES, 4)

        self.window = glfw.create_window(self.window_dims[0], self.window_dims[1], "Octrees!", None, None)
        if not self.window:
            glfw.terminate()
            return

        glfw.make_context_current(self.window)
        glfw.set_scroll_callback(self.window, self.scroll)
        gl.glClearColor(0.0, 0.6, 1.0, 0.0)

        self.compute_id = compile_compute_shader(cs)
        self.program_id = load_shaders(vs, fs)
        gl.glUseProgram(self.program_id)
        
        self.color_id = gl.glGetUniformLocation(self.program_id, "staticcolor")
        self.vertex_count_id = gl.glGetUniformLocation(self.compute_id, "vertexcount")

        self.vertex_array_id = gl.glGenVertexArrays(1)
        gl.glBindVertexArray(self.vertex_array_id)

        # create & load indexed vertex buffer
        self.vertex_buffer = gl.glGenBuffers(1)
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, self.vertex_buffer)
        indexed_vertex_data = numpy.array(indexed_vertecies, numpy.float32)
        gl.glBufferData(gl.GL_ARRAY_BUFFER, len(indexed_vertex_data)*4, indexed_vertex_data, gl.GL_STATIC_DRAW)

        # create & store cube index buffer
        self.cube_index_buffer = gl.glGenBuffers(1)
        gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, self.cube_index_buffer)
        cube_index_data = numpy.array(cube_indicies, numpy.uint16)
        gl.glBufferData(gl.GL_ELEMENT_ARRAY_BUFFER, len(cube_index_data)*2, cube_index_data, gl.GL_STATIC_DRAW)

        # create & store frame index buffer
        self.frame_index_buffer = gl.glGenBuffers(1)
        gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, self.frame_index_buffer)
        frame_index_data = numpy.array(frame_indicies, numpy.uint16)
        gl.glBufferData(gl.GL_ELEMENT_ARRAY_BUFFER, len(frame_index_data)*2, frame_index_data, gl.GL_STATIC_DRAW)

    
        # create & partially load transformation matricies 
        self.uniform_matricies = gl.glGenBuffers(1)
        gl.glBindBuffer(gl.GL_UNIFORM_BUFFER, self.uniform_matricies)
        gl.glBufferData(gl.GL_UNIFORM_BUFFER, 2 * glm.sizeof(glm.mat4), None, gl.GL_STATIC_DRAW)
        gl.glBufferSubData(gl.GL_UNIFORM_BUFFER, 0, glm.sizeof(glm.mat4), glm.value_ptr(self.projectionmatrix))
        
        # create & reserve space for barrier data
        self.barrier_vectors = gl.glGenBuffers(1)
        gl.glBindBuffer(gl.GL_UNIFORM_BUFFER, self.barrier_vectors)
        gl.glBufferData(gl.GL_UNIFORM_BUFFER, 2 * glm.sizeof(glm.vec4), None, gl.GL_STATIC_DRAW)

        # create indirect & matrix buffers
        self.matrix_buffer, self.indirect_buffer = gl.glGenBuffers(2)

        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, 0)
        gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, 0)
        gl.glBindBuffer(gl.GL_UNIFORM_BUFFER, 0)

        gl.glEnable(gl.GL_POLYGON_OFFSET_FILL)
        gl.glPolygonOffset(1, 0)
        gl.glLineWidth(1)
        gl.glEnable(gl.GL_LINE_SMOOTH)
        gl.glEnable(gl.GL_BLEND)
        gl.glEnable(gl.GL_DEPTH_TEST)
        gl.glDepthFunc(gl.GL_LEQUAL)
        gl.glEnable(gl.GL_CULL_FACE)


    def poll_octree_data(self):
        matricies = []
        for line in self.octree:
            if line["void"]:
                matricies.append(glm.translate(glm.mat4(1), glm.vec3(line["coords"])) * glm.scale(glm.mat4(1), glm.vec3(2**line["level"])))
        
        self.cube_count = len(matricies)
        matrix_array = glm.array(matricies)

        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, self.matrix_buffer)
        gl.glBufferData(gl.GL_ARRAY_BUFFER, matrix_array.nbytes, matrix_array.ptr, gl.GL_STATIC_DRAW)
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, 0)

        gl.glBindBuffer(gl.GL_DRAW_INDIRECT_BUFFER, self.indirect_buffer)
        gl.glBufferStorage(gl.GL_DRAW_INDIRECT_BUFFER, 32 * len(matricies), None, gl.GL_MAP_READ_BIT | gl.GL_MAP_WRITE_BIT)
        gl.glBindBuffer(gl.GL_DRAW_INDIRECT_BUFFER, 0)


    def draw(self):
        # Update UBO with new view matrix
        gl.glBindBuffer(gl.GL_UNIFORM_BUFFER, self.uniform_matricies)
        gl.glBufferSubData(gl.GL_UNIFORM_BUFFER, glm.sizeof(glm.mat4), glm.sizeof(glm.mat4), glm.value_ptr(self.probeKeyBoard()))

        # Update UBO with barrier data
        gl.glBindBuffer(gl.GL_UNIFORM_BUFFER, self.barrier_vectors)
        gl.glBufferSubData(gl.GL_UNIFORM_BUFFER, 0, glm.sizeof(glm.vec3), glm.value_ptr(self.direction))
        gl.glBufferSubData(gl.GL_UNIFORM_BUFFER, glm.sizeof(glm.vec4), glm.sizeof(glm.vec3), glm.value_ptr(self.barrier*self.direction + self.position))
        gl.glBindBuffer(gl.GL_UNIFORM_BUFFER, 0)

        # # Compute Commands
        gl.glUseProgram(self.compute_id)

        gl.glUniform1ui(self.vertex_count_id, len(cube_indicies))
        gl.glBindBufferBase(gl.GL_UNIFORM_BUFFER, 0, self.barrier_vectors)
        gl.glBindBufferBase(gl.GL_SHADER_STORAGE_BUFFER, 1, self.matrix_buffer)
        gl.glBindBufferBase(gl.GL_SHADER_STORAGE_BUFFER, 2, self.indirect_buffer)

        gl.glDispatchCompute(self.cube_count, 1, 1)
        gl.glMemoryBarrier(gl.GL_COMMAND_BARRIER_BIT | gl.GL_SHADER_STORAGE_BARRIER_BIT)

        gl.glUseProgram(self.program_id)
                
        # Bind VBO with verticies
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, self.vertex_buffer)
        gl.glEnableVertexAttribArray(0)
        gl.glVertexAttribPointer(0, 3, gl.GL_FLOAT, gl.GL_FALSE, 0, None)

        # Bind VBO with matricies and set attrib pointers
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, self.matrix_buffer)
    
        for i in range(4):
            gl.glEnableVertexAttribArray(2 + i)
            gl.glVertexAttribPointer(2 + i, 4, gl.GL_FLOAT, gl.GL_FALSE, glm.sizeof(glm.mat4), ctypes.c_void_p(glm.sizeof(glm.vec4) * i))
            gl.glVertexAttribDivisor(2 + i, 1)

        gl.glBindBufferBase(gl.GL_UNIFORM_BUFFER, 3, self.uniform_matricies)
        gl.glBindBuffer(gl.GL_DRAW_INDIRECT_BUFFER, self.indirect_buffer)

        # print(self.cube_count)
        # for i in range(self.cube_count):
        #     print(gl.glGetBufferSubData(gl.GL_DRAW_INDIRECT_BUFFER, 20 * i, 20))
        
        # sys.exit()
    
        # for i in [0, 2, 3, 4, 5]:
        #     print(gl.glGetVertexAttribfv(i, gl.GL_VERTEX_ATTRIB_RELATIVE_OFFSET))
    
        gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, self.cube_index_buffer)
        gl.glUniform3fv(self.color_id, 1, glm.value_ptr(self.fill_color))
        gl.glMultiDrawElementsIndirect(gl.GL_TRIANGLES, gl.GL_UNSIGNED_SHORT, None, self.cube_count, 20)
        
        gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, self.frame_index_buffer)
        gl.glUniform3fv(self.color_id, 1, glm.value_ptr(glm.vec3(1, 0, 0)))
        gl.glMultiDrawElementsIndirect(gl.GL_LINES, gl.GL_UNSIGNED_SHORT, None, self.cube_count, 20)

        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, 0)
        gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, 0)
        gl.glBindBuffer(gl.GL_DRAW_INDIRECT_BUFFER, 0)


    def main(self):
        self.poll_octree_data()
        while not glfw.get_key(self.window, glfw.KEY_ESCAPE) == glfw.PRESS:
            gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)

            self.draw()
            glfw.swap_buffers(self.window)
            glfw.poll_events() 

        glfw.terminate()


    def scroll(self, window, xoffset, yoffset):
        if glfw.get_key(self.window, glfw.KEY_LEFT_ALT) == glfw.PRESS:
            self.barrier = (self.barrier + yoffset) if self.barrier >= 0 else 0
        else:
            self.dist = (self.dist + yoffset) if self.dist >= 0 else 0


    def on_key(self, window, key, scancode, action, mods):...


    def probeKeyBoard(self):  
        """returns updated view matrix"""
        self.current_time = glfw.get_time()
        self.delta_time = self.current_time - self.last_time
        self.last_time = self.current_time

        print("\r                                                 \r", round(self.delta_time, 5), "\t|", 1/self.delta_time, sep="", end="")

        if glfw.get_key(self.window, glfw.KEY_W) == glfw.PRESS:
            self.v_angle -= (self.delta_time * self.speed) if (self.v_angle > math.pi/-2) else 0

        if glfw.get_key(self.window, glfw.KEY_S) == glfw.PRESS:
            self.v_angle += (self.delta_time * self.speed) if (self.v_angle < math.pi/2) else 0

        if glfw.get_key(self.window, glfw.KEY_D) == glfw.PRESS:
            self.h_angle += self.delta_time * self.speed

        if glfw.get_key(self.window, glfw.KEY_A) == glfw.PRESS:
            self.h_angle -= self.delta_time * self.speed

        if glfw.get_mouse_button(self.window, glfw.MOUSE_BUTTON_LEFT) == glfw.PRESS:
            mouse_x, mouse_y = glfw.get_cursor_pos(self.window)
            if self.last_mouse_x and self.last_mouse_y:
                self.h_angle += (self.last_mouse_x - mouse_x) / self.window_dims[0] * 4
                self.v_angle += (self.last_mouse_y - mouse_y) / self.window_dims[1] * 4
            self.last_mouse_x, self.last_mouse_y = mouse_x, mouse_y
        elif self.last_mouse_x or self.last_mouse_y:
            self.last_mouse_x, self.last_mouse_y = None, None

        self.direction = glm.vec3(
                            math.cos(self.v_angle) * math.sin(self.h_angle),
                            math.sin(self.v_angle),
                            math.cos(self.v_angle) * math.cos(self.h_angle))

        right = glm.vec3(
                        math.sin(self.h_angle - math.pi/2),
                        0,
                        math.cos(self.h_angle - math.pi/2))

        up = glm.cross(right, self.direction)

        self.position = self.fulcrum - self.dist * self.direction
        return glm.lookAt(self.position, self.fulcrum, up)



from io import BufferedWriter
from queue import LifoQueue
from struct import pack
  


def scanner(octree):
    def process(data):
        nonlocal lastlevel
        for i in range(lastlevel - data["level"]):
            counter.put(0)
            lastlevel -= 1
            yield Command.create_node()
        
        yield Command.fill_node(data["data"])
        lastlevel += increment_counter()
        
    def increment_counter():
        if counter.empty():
            return 0 
        lastcount = counter.get()
        lastcount += 1
        if lastcount == 8:
            counter.task_done()
            return 1 + increment_counter()
        else:
            counter.put(lastcount)
            return 0

    lastlevel = octree.level
    counter = LifoQueue()
    yield Command.header()
    yield Command.seed(octree.level)
    command:Command = None
    for data in iter(octree):
        for nxtcmd in process(data):
            if command is None:
                command = nxtcmd
            elif nxtcmd == command:
                command.count += 1
            else:
                yield command
                command = nxtcmd
    yield command


class Command:
    singletons = {
        None:"nlnd",
        Ellipsis:"vdnd",
        False:"fsnd",
        True:"trnd"
    }

    def __init__(self, cmd, value=None):
        self.cmd = cmd
        self.count = 1
        self.value = value

    def __str__(self):
        out = self.cmd
        if self.value is not None:
            out += " " + str(self.value)
        return out

    def __eq__(self, other):
        if type(other) == type(self):
            if other.cmd == self.cmd:
                if other.value == self.value:
                    return True
        return False

    @classmethod
    def header(cls, version="0.0.2"):
        return cls("header", version)

    @classmethod
    def seed(cls, level):
        return cls("seed", level)

    @classmethod
    def fill_node(cls, value):
        if value in cls.singletons:
            return cls(cls.singletons[value])
        else:
            return cls("flnd", value)

    @classmethod
    def create_node(cls):
        return cls("crnd")


class SavingStream:
    types = {
          'i8':b'\x20',  'i16':b'\x21',  'i32':b'\x22',  'i64':b'\x23', 
          'u8':b'\x24',  'u16':b'\x25',  'u32':b'\x26',  'u64':b"\x27", 
         'f32':b'\x28',  'f64':b'\x29',  'c64':b'\x2a', 'c128':b'\x2b', 
        "flse":b"\x2c", "true":b"\x2d", 
         'Str':b'\x40', 'List':b'\x41', 'Dict':b'\x42',  'Set':b'\x43'
    }
    py_standard = {
        int: "i32", 
        float: "f64",
        complex: "c64",
        str: "Str",
        list: "List",
        dict: "Dict",
        set: "Set"
    }

    
    def __init__(self, io:BufferedWriter) -> None:
        self.io = io


    def convert(self, value):
        type_byte = self.py_standard[type(value)]
        self.io.write(self.types[type_byte])
        getattr(self, type_byte)(value)


    def write(self, value:bytes):
        self.io.write(value)


    def i8(self, value:int):
        self.io.write(value.to_bytes(1, byteorder="big", signed=True))

    def i16(self, value:int):
        self.io.write(value.to_bytes(2, byteorder="big", signed=True))

    def i32(self, value:int):
        self.io.write(value.to_bytes(4, byteorder="big", signed=True))

    def i64(self, value:int):
        self.io.write(value.to_bytes(8, byteorder="big", signed=True))

    def u8(self, value:int):
        self.io.write(value.to_bytes(1, byteorder="big", signed=False))
    
    def u16(self, value:int):
        self.io.write(value.to_bytes(2, byteorder="big", signed=False))

    def u32(self, value:int):
        self.io.write(value.to_bytes(4, byteorder="big", signed=False))

    def u64(self, value:int):
        self.io.write(value.to_bytes(8, byteorder="big", signed=False))

    def f32(self, value:float):
        self.io.write(pack("f", value))

    def f64(self, value:float):
        self.io.write(pack("d", value))

    def c64(self, value:complex):
        self.io.write(pack("ff", value.real, value.imag))

    def c128(self, value:complex):
        self.io.write(pack("dd", value.real, value.imag))

    def Str(self, value:str):
        self.u16(len(value))
        self.io.write(value.encode())

    def List(self, value:list):
        self.u16(len(value))
        for item in value:
            self.convert(item) 

    def Dict(self, value:dict):
        self.u16(len(value))
        for key, value in value.items():
            self.convert(key); self.convert(value) 

    def Set(self, value:set):
        self.u16(len(value))
        for item in value:
            self.convert(item) 


class Saver:
    conversion = {
        'header':b'\x00', 'seed':b'\x01', 

        'crnd':b'\x04', 'flnd':b'\x07', 
        'nlnd':b'\x08', 'vdnd':b'\x09', 'fsnd':b'\x0a', 'trnd':b'\x0b', 
    }


    def __init__(self, io:BufferedWriter, factory) -> None:
        self.factory = factory
        self.converter = SavingStream(io)


    def translate(self, command:Command):
        """Routes the incoming commands that a octree decomposes into towards the individual transcription commands"""
        getattr(self, command.cmd)(command)
    

    def header(self, command:Command):
        self.converter.write(self.conversion[command.cmd])
        self.converter.convert(command.value)

    def seed(self, command:Command):
        self.converter.write(self.conversion[command.cmd])
        self.converter.convert(command.value)

    def crnd(self, command:Command):
        self.converter.write(self.conversion[command.cmd])
        self.converter.u8(command.count)

    def flnd(self, command:Command):
        self.converter.write(self.conversion[command.cmd])
        self.converter.u8(command.count)
        self.factory.to_file(command.value, self.converter)
    
    def nlnd(self, command:Command):
        self.converter.write(self.conversion[command.cmd])
        self.converter.u8(command.count)

    def vdnd(self, command:Command):
        self.converter.write(self.conversion[command.cmd])
        self.converter.u8(command.count)

    def fsnd(self, command:Command):
        self.converter.write(self.conversion[command.cmd])
        self.converter.u8(command.count)

    def trnd(self, command:Command):
        self.converter.write(self.conversion[command.cmd])
        self.converter.u8(command.count)





from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()
desc = (here / "README.md").read_text()

setup(
    name="regOct",
    version='1.1.2',
    description="A module implementing octrees for python",
    long_description=desc,
    long_description_content_type='text/markdown',
    url='https://github.com/runoverPy/regOct',
    author="runover",
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.0",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    packages=find_packages(where="src"),
    package_dir = {"":"src"},
    python_requires=">=3, <4",
    install_requires=[
        "PyGLM",
        "numpy",
        "keyboard",
        "glfw",
        "PyOpenGL",
        "tqdm",
    ],
)